@@ -1,9 +1,9 @@
 #if !defined(lint) && !defined(SABER)
-static char rcsid[] = "$Id: ns_lexer.c,v 8.10 1997/06/09 17:46:56 halley Exp $";
+static char rcsid[] = "$Id: ns_lexer.c,v 8.10 1999/03/16 02:21:29 vixie Exp $";
 #endif /* not lint */
 
 /*
- * Copyright (c) 1996, 1997 by Internet Software Consortium.
+ * Copyright (c) 1996-1999 by Internet Software Consortium.
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,6 +22,8 @@ static char rcsid[] = "$Id: ns_lexer.c,v 8.10 1997/06/09 17:46:56 halley Exp $";
 #include "port_before.h"
 
 #include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 
 #include <netinet/in.h>
 #include <arpa/nameser.h>
@@ -38,6 +40,7 @@ static char rcsid[] = "$Id: ns_lexer.c,v 8.10 1997/06/09 17:46:56 halley Exp $";
 
 #include <isc/eventlib.h>
 #include <isc/logging.h>
+#include <isc/memcluster.h>
 
 #include "port_after.h"
 
@@ -51,6 +54,7 @@ typedef enum lexer_state {
 } LexerState;
 
 #define LEX_EOF			0x01
+#define LEXER_MAX_PUSHBACK	2
 
 typedef struct lexer_file_context {
 	const char *	name;
@@ -60,6 +64,8 @@ typedef struct lexer_file_context {
 	u_int		flags;
         int		warnings;
         int		errors;
+	u_int		pushback_count;
+	char		pushback[LEXER_MAX_PUSHBACK];
 	struct lexer_file_context *
 			next;
 } *LexerFileContext;
@@ -88,7 +94,7 @@ static char special_chars[256];
 static int last_token;
 static YYSTYPE last_yylval;
 
-static int init_once_done = 0;
+static int lexer_initialized = 0;
 
 /*
  * Problem Reporting
@@ -215,20 +221,26 @@ static struct keyword keywords[] = {
 	{"acl", T_ACL}, 
 	{"address", T_ADDRESS},
 	{"algorithm", T_ALGID},
+	{"allow", T_ALLOW},
 	{"allow-query", T_ALLOW_QUERY}, 
 	{"allow-transfer", T_ALLOW_TRANSFER},
 	{"allow-update", T_ALLOW_UPDATE},
 	{"also-notify", T_ALSO_NOTIFY},
 	{"auth-nxdomain", T_AUTH_NXDOMAIN},
+	{"blackhole", T_BLACKHOLE},
 	{"bogus", T_BOGUS},
 	{"category", T_CATEGORY},
+	{"class", T_CLASS},
 	{"channel", T_CHANNEL},
 	{"check-names", T_CHECK_NAMES},
 	{"cleaning-interval", T_CLEAN_INTERVAL},
+	{"controls", T_CONTROLS},
 	{"coresize", T_CORESIZE},
 	{"datasize", T_DATASIZE},
+	{"deallocate-on-exit", T_DEALLOC_ON_EXIT},
 	{"debug", T_DEBUG},
 	{"default", T_DEFAULT},
+	{"dialup", T_DIALUP},
 	{"directory", T_DIRECTORY}, 
 	{"dump-file", T_DUMP_FILE},
 	{"dynamic", T_DYNAMIC},
@@ -241,34 +253,53 @@ static struct keyword keywords[] = {
 	{"first", T_FIRST}, 
 	{"forward", T_FORWARD},
 	{"forwarders", T_FORWARDERS},
+	{"group", T_GROUP},
+	{"has-old-clients", T_HAS_OLD_CLIENTS},
+	{"heartbeat-interval", T_HEARTBEAT},
 	{"hint", T_HINT},
+	{"host-statistics", T_HOSTSTATS},
 	{"if-no-answer", T_IF_NO_ANSWER},
 	{"if-no-domain", T_IF_NO_DOMAIN},
 	{"ignore", T_IGNORE},
 	{"include", T_INCLUDE},
+	{"inet", T_INET},
 	{"interface-interval", T_INTERFACE_INTERVAL},
+	{"ixfr-base", T_FILE_IXFR},
+	{"ixfr-tmp-file", T_IXFR_TMP},
 	{"key", T_SEC_KEY},
 	{"keys", T_KEYS},
+	{"lame-ttl", T_LAME_TTL},
 	{"listen-on", T_LISTEN_ON},
 	{"logging", T_LOGGING},
+	{"maintain-ixfr-base", T_MAINTAIN_IXFR_BASE},
 	{"many-answers", T_MANY_ANSWERS},
 	{"master", T_MASTER},
 	{"masters", T_MASTERS},
+	{"max-ixfr-log-size", T_MAX_LOG_SIZE_IXFR},
+	{"max-ncache-ttl", T_MAX_NCACHE_TTL},
 	{"max-transfer-time-in", T_MAX_TRANSFER_TIME_IN},
+	{"memstatistics-file", T_MEMSTATS_FILE},
 	{"multiple-cnames", T_MULTIPLE_CNAMES},
+	{"name", T_NAME},
 	{"named-xfer", T_NAMED_XFER},
 	{"no", T_NO},
 	{"notify", T_NOTIFY},
 	{"null", T_NULL_OUTPUT},
 	{"one-answer", T_ONE_ANSWER},
 	{"only", T_ONLY},
+	{"order", T_ORDER},
 	{"options", T_OPTIONS},
+	{"owner", T_OWNER},
+	{"perm", T_PERM},
 	{"pid-file", T_PIDFILE},
 	{"port", T_PORT},
 	{"print-category", T_PRINT_CATEGORY},
 	{"print-severity", T_PRINT_SEVERITY},
 	{"print-time", T_PRINT_TIME},
+	{"pubkey", T_PUBKEY},
 	{"query-source", T_QUERY_SOURCE},
+	{"rfc2308-type1", T_RFC2308_TYPE1},
+	{"rrset-order", T_RRSET_ORDER},
 	{"recursion", T_RECURSION},
 	{"response", T_RESPONSE},
 	{"secret", T_SECRET},
@@ -276,20 +307,30 @@ static struct keyword keywords[] = {
 	{"severity", T_SEVERITY}, 
 	{"size", T_SIZE}, 
 	{"slave", T_SLAVE},
+#ifdef SORT_RESPONSE
+	{"sortlist", T_SORTLIST},
+#endif /* SORT_RESPONSE */
 	{"stacksize", T_STACKSIZE},
 	{"statistics-file", T_STATS_FILE},
 	{"statistics-interval", T_STATS_INTERVAL},
 	{"stub", T_STUB},
+	{"support-ixfr", T_SUPPORT_IXFR},
 	{"syslog", T_SYSLOG}, 
 	{"topology", T_TOPOLOGY},
 	{"transfer-format", T_TRANSFER_FORMAT}, 
+	{"transfer-source", T_TRANSFER_SOURCE},
 	{"transfers", T_TRANSFERS}, 
 	{"transfers-in", T_TRANSFERS_IN}, 
 	{"transfers-out", T_TRANSFERS_OUT}, 
 	{"transfers-per-ns", T_TRANSFERS_PER_NS}, 
 	{"true", T_TRUE}, 
+	{"trusted-keys", T_TRUSTED_KEYS},
 	{"type", T_TYPE},
+	{"unix", T_UNIX},
 	{"unlimited", T_UNLIMITED},
+	{"use-id-pool", T_USE_ID_POOL},
+	{"use-ixfr", T_USE_IXFR},
+	{"version", T_VERSION},
 	{"versions", T_VERSIONS}, 
 	{"warn", T_WARN},
 	{"yes", T_YES}, 
@@ -335,9 +376,9 @@ lexer_begin_file(const char *filename, FILE *stream) {
 			return;
 		}
 	}
-	lf = (LexerFileContext)malloc(sizeof (struct lexer_file_context));
+	lf = (LexerFileContext)memget(sizeof (struct lexer_file_context));
 	if (lf == NULL)
-		panic("malloc failed in lexer_begin_file", NULL);
+		panic("memget failed in lexer_begin_file", NULL);
 	INSIST(stream != NULL);
 	lf->stream = stream;
 	lf->name = filename;  /* note copy by reference */
@@ -346,6 +387,7 @@ lexer_begin_file(const char *filename, FILE *stream) {
 	lf->flags = 0;
 	lf->warnings = 0;
 	lf->errors = 0;
+	lf->pushback_count = 0;
 	lf->next = current_file;
 	current_file = lf;
 }
@@ -358,21 +400,36 @@ lexer_end_file(void) {
 	lf = current_file;
 	current_file = lf->next;
 	fclose(lf->stream);
-	free(lf);
+	memput(lf, sizeof *lf);
 }
 
 /*
  * Character Input
  */
 
+#define LEXER_GETC(c, cf) \
+	do { \
+		if ((cf)->pushback_count > 0) { \
+			(cf)->pushback_count--; \
+			(c) = (cf)->pushback[(cf)->pushback_count]; \
+		} else \
+			(c) = getc((cf)->stream); \
+	} while (0);
+
+#define LEXER_UNGETC(c, cf) \
+	do { \
+		INSIST((cf)->pushback_count < LEXER_MAX_PUSHBACK); \
+		(cf)->pushback[(cf)->pushback_count++] = (c); \
+	} while (0);
+
 static void
 scan_to_comment_end(int c_plus_plus_style) {
-	int c, nc;
+	int c;
 	int done = 0;
 	int prev_was_star = 0;
 
 	while (!done) {
-		c = getc(current_file->stream);
+		LEXER_GETC(c, current_file);
 		switch (c) {
 		case EOF:
 			if (!c_plus_plus_style)
@@ -394,7 +451,7 @@ scan_to_comment_end(int c_plus_plus_style) {
 				   we want it to be a delimiter for
 				   anything before the comment
 				   started */
-				ungetc(c, current_file->stream);
+				LEXER_UNGETC(c, current_file);
 				done = 1;
 			} else {
 				current_file->line_number++;
@@ -414,7 +471,7 @@ get_next_char(int comment_ok) {
 	if (current_file->flags & LEX_EOF)
 		return (EOF);
 
-	c = getc(current_file->stream);
+	LEXER_GETC(c, current_file);
 
 	if (comment_ok) {
 		while (c == '/' || c == '#') {
@@ -422,9 +479,9 @@ get_next_char(int comment_ok) {
 				scan_to_comment_end(1);
 				if (current_file->flags & LEX_EOF)
 					return (EOF);
-				c = getc(current_file->stream);
+				LEXER_GETC(c, current_file);
 			} else {
-				nc = getc(current_file->stream);
+				LEXER_GETC(nc, current_file);
 				switch (nc) {
 				case EOF:
 					current_file->flags |= LEX_EOF;
@@ -434,10 +491,10 @@ get_next_char(int comment_ok) {
 					scan_to_comment_end((nc == '/'));
 					if (current_file->flags & LEX_EOF)
 						return (EOF);
-					c = getc(current_file->stream);
+					LEXER_GETC(c, current_file);
 					break;
 				default:
-					ungetc((nc), current_file->stream);
+					LEXER_UNGETC(nc, current_file);
 					return ('/');
 				}
 			}
@@ -456,7 +513,7 @@ put_back_char(int c) {
 	if (c == EOF)
 		current_file->flags |= LEX_EOF;
 	else {
-		ungetc((c), current_file->stream);
+		LEXER_UNGETC(c, current_file);
 		if (c == '\n')
 			current_file->line_number--;
 	}
@@ -480,9 +537,7 @@ dup_identifier(LexerIdentifier id) {
 	char *duplicate;
 
 	INSIST(id != NULL);
-	duplicate = strdup(id->buffer);
-	if (duplicate == NULL)
-		panic("strdup failed in dup_identifier", NULL);
+	duplicate = savestr(id->buffer, 1);
 	return (duplicate);
 }
 
@@ -523,7 +578,7 @@ add_to_identifier(LexerIdentifier id, int c) {
  */
 int
 yylex() {
-	int c, i;
+	int c;
 	int comment_ok = 1;
 	int token = -1;
 	symbol_value value;
@@ -587,7 +642,7 @@ yylex() {
 				put_back_char(c);
 				current_file->state = scan;
 				finish_identifier(id);
-				yylval.num = atoi(id->buffer);
+				yylval.num = strtol(id->buffer, (char**)0, 0);
 				token = L_NUMBER;
 			}
 			break;
@@ -710,8 +765,8 @@ import_all_constants() {
 	import_constants(syslog_constants, SYM_SYSLOG);
 }
 
-static
-void init_once() {
+void
+lexer_initialize() {
 	memset(special_chars, 0, sizeof special_chars);
 	special_chars[';'] = 1;
 	special_chars['{'] = 1;
@@ -720,18 +775,28 @@ void init_once() {
 	special_chars['/'] = 1;
 	special_chars['"'] = 1;
 	special_chars['*'] = 1;
-	id = (LexerIdentifier)malloc(sizeof (struct lexer_identifier));
+	id = (LexerIdentifier)memget(sizeof (struct lexer_identifier));
 	if (id == NULL)
-		panic("malloc failed in init_once", NULL);
+		panic("memget failed in lexer_initialize", NULL);
 	init_keywords();
 	import_all_constants();
-	init_once_done = 1;
+	lexer_initialized = 1;
 }
 
 void
-lexer_initialize(void) {
-	if (!init_once_done)
-		init_once();
+lexer_setup(void) {
+	REQUIRE(lexer_initialized);
+
 	current_file = NULL;    /* XXX should we INSIST(current_file==NULL)? */
 	INSIST(id != NULL);
 }
+
+void
+lexer_shutdown(void) {
+	REQUIRE(lexer_initialized);
+
+	free_symbol_table(keyword_table);
+	free_symbol_table(constants);
+	memput(id, sizeof (struct lexer_identifier));
+	lexer_initialized = 0;
+}
